/*
 *	interrup.c - interrupt handling
 *	FF sometime in 1994-1997
 */

/*
Parts of this file copyright 1994-1997 Fabrice Francs.
Parts of this file copyright 2000-2003 Andr Majorel.

This program is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published by the
Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef __GO32__

#include <dpmi.h>
extern void Timer_Int();
extern void Keyboard_Int();
extern char delayed_key_flag;
extern void Serial_Int();
extern void Enable_Serial_Int();
extern int Read_Kbd_Buf();
extern char disk;
extern char serial_name[];

extern int Com_port;

_go32_dpmi_seginfo old_keyboard, new_keyboard;
_go32_dpmi_seginfo old_real_keyboard, new_real_keyboard;
_go32_dpmi_seginfo old_timer, new_timer;
_go32_dpmi_seginfo old_serial, new_serial;
_go32_dpmi_seginfo old_real_serial, new_real_serial;
_go32_dpmi_registers regs;

poll_keyboard()
{
	int scancode;
	if (0 /*ProgrammableKbd*/) {
		if (delayed_key_flag) make_delayed_key();
		else {
			scancode=Read_Kbd_Buf();
			if (scancode) keyboard_handler(scancode);
		}
	} else while(scancode=Read_Kbd_Buf()) keyboard_handler(scancode);
}

hook_interrupts()
{
	new_timer.pm_offset = (int)Timer_Int;
	_go32_dpmi_allocate_iret_wrapper(&new_timer);
	new_timer.pm_selector = _go32_my_cs();
	_go32_dpmi_get_protected_mode_interrupt_vector(8,&old_timer);
	_go32_dpmi_set_protected_mode_interrupt_vector(8,&new_timer);

	new_keyboard.pm_offset = (int)Keyboard_Int;
	_go32_dpmi_allocate_iret_wrapper(&new_keyboard);
	new_keyboard.pm_selector = _go32_my_cs();
	_go32_dpmi_get_protected_mode_interrupt_vector(9,&old_keyboard);
	_go32_dpmi_set_protected_mode_interrupt_vector(9,&new_keyboard);

	new_real_keyboard.pm_offset = (int)Keyboard_Int;
	_go32_dpmi_allocate_real_mode_callback_iret(&new_real_keyboard,&regs);
	_go32_dpmi_get_real_mode_interrupt_vector(9,&old_real_keyboard);
	_go32_dpmi_set_real_mode_interrupt_vector(9,&new_real_keyboard);

    if (disk) {
	new_serial.pm_offset = (int)Serial_Int;
	_go32_dpmi_allocate_iret_wrapper(&new_serial);
	new_serial.pm_selector = _go32_my_cs();
	_go32_dpmi_get_protected_mode_interrupt_vector(12-Com_port,&old_serial);
	_go32_dpmi_set_protected_mode_interrupt_vector(12-Com_port,&new_serial);

	new_real_serial.pm_offset = (int)Serial_Int;
	_go32_dpmi_allocate_real_mode_callback_iret(&new_real_serial,&regs);
	_go32_dpmi_get_real_mode_interrupt_vector(12-Com_port,&old_real_serial);
	_go32_dpmi_set_real_mode_interrupt_vector(12-Com_port,&new_real_serial);

	Enable_Serial_Int();
    }
}

restore_interrupts()
{
	_go32_dpmi_set_protected_mode_interrupt_vector(8,&old_timer);
	_go32_dpmi_set_protected_mode_interrupt_vector(9,&old_keyboard);
	_go32_dpmi_free_iret_wrapper(&new_keyboard);
	_go32_dpmi_set_real_mode_interrupt_vector(9,&old_real_keyboard);
	_go32_dpmi_free_real_mode_callback(&new_real_keyboard);
    if (disk) {
	_go32_dpmi_set_protected_mode_interrupt_vector(12-Com_port,&old_serial);
	_go32_dpmi_free_iret_wrapper(&new_serial);
	_go32_dpmi_set_real_mode_interrupt_vector(12-Com_port,&old_real_serial);
	_go32_dpmi_free_real_mode_callback(&new_real_serial);
    }
}

#else

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/time.h>
#ifdef X11
#include <X11/Xlib.h>
#else
#include <termios.h>
#include <linux/kd.h>
#endif

#include "xeuphoric.h"
#include "stats.h"
#include "time.h"

#ifdef X11
extern void keyboard_handler(XEvent *ev);
extern void x11_trace(int);
#else
extern void keyboard_handler(unsigned scan);
#endif
void poll_keyboard(void);
extern char delayed_key_flag;
extern int serial;
extern int Com_port;
extern int disk;
extern char serial_name[];
extern int xputimage_in_progress;

static struct sigaction oldsiga_segv, oldsiga_kill;
static struct itimerval itimer1, itimer2;
#ifndef X11
static struct termios oldkbdtermios, newkbdtermios;
static int oldkbmode;
static int kbd_fd;
static unsigned char scancode;
#endif

extern unsigned char Tick;
void alarm_handler()
{
	synchronize();
	/* KLUDGE if XPutImage() is in progress, don't poll the
	   keyboard, Xlib is not sufficiently reentrant for that. */
	stats.polls++;
	if (! xputimage_in_progress)
	  poll_keyboard();
	signal(SIGALRM,alarm_handler);
}

void close_keyboard() {
#ifdef X11
	;
#else
	ioctl(kbd_fd, KDSKBMODE, oldkbmode);
	tcsetattr(kbd_fd, 0, &oldkbdtermios);
	close(kbd_fd);
#endif
	sigaction(SIGSEGV, &oldsiga_segv, NULL);
	sigaction(SIGKILL, &oldsiga_kill, NULL);
}

static void handle_segv() {
	close_keyboard();
	sigaction(SIGSEGV, &oldsiga_segv, NULL);
	raise(SIGSEGV);
}

static void handle_kill() {
	close_keyboard();
	sigaction(SIGKILL, &oldsiga_kill, NULL);
	raise(SIGKILL);
}

void open_keyboard(void)
{
	struct sigaction siga;
#ifndef X11
/*
 * Keyboard I/O based on showkey.c from kbd-0.84 package.
 */

	if ((kbd_fd = open("/dev/console", O_RDONLY)) < 0) {
		printf("Cannot open /dev/console.\n");
		return -1;
	}
	if (ioctl(kbd_fd, KDGKBMODE, &oldkbmode)) {
		printf("Cannot get keyboard mode.\n");
		return -1;
	}

	tcgetattr(kbd_fd, &oldkbdtermios);
	newkbdtermios = oldkbdtermios;

	newkbdtermios.c_lflag = newkbdtermios.c_lflag & ~ (ICANON | ECHO | ISIG);
	newkbdtermios.c_cc[VMIN] = 0;   /* Making these 0 seems to have the */
	newkbdtermios.c_cc[VTIME] = 0;  /* desired effect. */

	tcsetattr(kbd_fd, TCSAFLUSH, &newkbdtermios);

	ioctl(kbd_fd, KDSKBMODE, K_RAW);
#endif
	siga.sa_handler = handle_segv;
	siga.sa_flags = 0;
	memset(&siga.sa_mask, 0, sizeof siga.sa_mask);
	sigaction(SIGSEGV, &siga, &oldsiga_segv);
	siga.sa_handler = handle_kill;
	sigaction(SIGKILL, &siga, &oldsiga_kill);
}

void poll_keyboard(void)
{
#ifdef X11
    XEvent ev;
    while (x11_key(&ev)) {
	keyboard_handler(&ev);
    }
#else
    if (0 /*ProgrammableKbd*/) {
	if (delayed_key_flag) make_delayed_key();
	else while (!delayed_key_flag && read(kbd_fd,&scancode,1)==1)
		keyboard_handler(scancode);
    } else while (read(kbd_fd,&scancode,1)==1) keyboard_handler(scancode);
#endif
}


void hook_interrupts(void)
{
#ifndef X11
	system("stty raw");
#endif
	open_keyboard();
	signal(SIGALRM,alarm_handler);
	itimer1.it_interval.tv_usec=itimer1.it_value.tv_usec=20000L;
	setitimer(ITIMER_REAL,&itimer1,&itimer2);
}

void restore_interrupts(void)
{
	itimer1.it_interval.tv_usec=itimer1.it_value.tv_usec=0L;
	setitimer(ITIMER_REAL,&itimer1,&itimer2);
	close_keyboard();
}

#endif
