/*
 *	keyb_us.c - keyboard input for Xeuphoric
 *	AYM 2002-01-22
 */

/*
This file is copyright Andr Majorel 2002.

This program is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published by the
Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307, USA.
*/


#include <limits.h>
#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "xeuphoric.h"
#include "keyb_us.h"
#include "screenshot.h"

extern short Sys_Request;
extern long  cycles;
extern long  initial_cycles;

char env_screen;
char exit_menu;
unsigned char Kbd_Matrix[65];		/* FIXME should be 8 */


/* XFree86 keycodes in more or less human-readable form */
typedef enum
{
  XKC_ESC = 0x09,
  XKC_F1  = 0x43,
  XKC_F2  = 0x44,
  XKC_F3  = 0x45,
  XKC_F4  = 0x46,
  XKC_F5  = 0x47,
  XKC_F6  = 0x48,
  XKC_F7  = 0x49,
  XKC_F8  = 0x4a,
  XKC_F9  = 0x4b,
  XKC_F10 = 0x4c,
  XKC_F11 = 0x5f,
  XKC_F12 = 0x60,

  XKC_PRT = 0x6f,
  XKC_SCR = 0x4e,
  XKC_PAU = 0x6e,

  XKC_1   = 0x0a,
  XKC_2   = 0X0b,
  XKC_3   = 0x0c,
  XKC_4   = 0x0d,
  XKC_5   = 0x0e,
  XKC_6   = 0x0f,
  XKC_7   = 0x10,
  XKC_8   = 0x11,
  XKC_9   = 0x12,
  XKC_0   = 0x13,
  XKC_MIN = 0x14,			/* Minus */
  XKC_EQ  = 0x15,			/* Equal sign */
  XKC_BSL = 0x33,			/* Backslash */
  XKC_BS  = 0x16,			/* Backspace */
  XKC_TAB = 0x17,
  XKC_CAP = 0x42,			/* Caps lock */
  XKC_Q   = 0x18,
  XKC_W   = 0x19,
  XKC_E   = 0x1a,
  XKC_R   = 0x1b,
  XKC_T   = 0x1c,
  XKC_Y   = 0x1d,
  XKC_U   = 0x1e,
  XKC_I   = 0x1f,
  XKC_O   = 0x20,
  XKC_P   = 0x21,
  XKC_BRL = 0x22,			/* Left square bracket */
  XKC_BRR = 0x23,			/* Right square bracket */
  XKC_ENT = 0x24,
  XKC_A   = 0x26,
  XKC_S   = 0x27,
  XKC_D   = 0x28,
  XKC_F   = 0x29,
  XKC_G   = 0x2a,
  XKC_H   = 0x2b,
  XKC_J   = 0x2c,
  XKC_K   = 0x2d,
  XKC_L   = 0x2e,
  XKC_SEM = 0x2f,			/* Semicolon */
  XKC_APO = 0x30,			/* Apostrophe */
  XKC_SHL = 0x32,			/* Left shift */
  XKC_LES = 0x5e,			/* Lesser-than (not on all keyb.) */
  XKC_Z   = 0x34,
  XKC_X   = 0x35,
  XKC_C   = 0x36,
  XKC_V   = 0x37,
  XKC_B   = 0x38,
  XKC_N   = 0x39,
  XKC_M   = 0x3a,
  XKC_COM = 0x3b,			/* Comma */
  XKC_FUL = 0x3c,			/* Full stop */
  XKC_SLA = 0x3d,			/* Slash */
  XKC_SHR = 0x3e,			/* Right shift */
  XKC_CTL = 0x25,			/* Left ctrl */
  XKC_ALL = 0x40,			/* Left alt */
  XKC_SPC = 0x41,			/* Space */
  XKC_ALR = 0x71,			/* Right alt */
  XKC_CTR = 0x6d,			/* Right ctrl */

  XKC_INS = 0x6a,
  XKC_HOM = 0x61,
  XKC_PU  = 0x63,
  XKC_DEL = 0x6b,
  XKC_END = 0x67,
  XKC_PD  = 0x69,

  XKC_UP  = 0x62,
  XKC_LEF = 0x64,
  XKC_DOW = 0x68,
  XKC_RIG = 0x66,

  XKC_NUM = 0x4d,
  XKC_PSL = 0x70,
  XKC_PAS = 0x3f,
  XKC_PMI = 0x52,
  XKC_P7  = 0x4f,
  XKC_P8  = 0x50,
  XKC_P9  = 0x51,
  XKC_PPL = 0x56,
  XKC_P4  = 0x53,
  XKC_P5  = 0x54,
  XKC_P6  = 0x55,
  XKC_P1  = 0x57,
  XKC_P2  = 0x58,
  XKC_P3  = 0x59,
  XKC_PEN = 0x6c,
  XKC_P0  = 0x5a,
  XKC_PFS = 0x5b
} x_keycode_t;


#if 0
/* The index is the XFree86 keycode (for a PC keyboard). The
   value is the Oric scancode. */
static signed char keycode_to_oric[] =
{
   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, /*   0:                          */
   -1, 051, 050, 062, 070, 032, 020, 012, /*   8: ?? Esc1  2  3  4  5  6   */
  000, 007, 013, 027, 033, 077, 055, 067, /*  16: 7  8  9  0  )  =  BS Tab */
  061, 076, 036, 021, 011, 006, 005, 015, /*  24: Q  W  E  R  T  Y  U  I   */
  025, 035, 075, 065, 057, 042, 056, 066, /*  32: O  P  [  ]  RetCtlA  S   */
  071, 031, 026, 016, 001, 003, 017, 023, /*  40: D  F  G  H  J  K  L  ;   */
  073, 067, 044, 063, 052, 060, 072, 030, /*  48: '    Shf\  Z  X  C  V   */
  022, 010, 002, 014, 024, 037, 047, 045, /*  56: B  N  M  .  ,  /  ShfAlt */
  045, 004, 053,  -2,  -2,  -2,  -2,  -2, /*  64: AltSp CapF1 F2 F3 F4 F5  */
   -2,  -2,  -2,  -2,  -2,  -1,  -1,  -1, /*  72: F6 F7 F8 F9 F10NumScr7   */
  034,  -1,  -1, 054,  -1, 074,  -1,  -1, /*  80: 8  9  -  4  5  6  +  1   */
  064,  -1,  -1,  -1,  -1,  -1,  -1,  -1, /*  88: 2  3  0  .  Sys?  <  F11 */
   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, /*  96: ?? ?? ?? ?? ?? ?? ?? ??  */
   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, /* 104: ?? ?? ?? ?? ?? ?? ?? ??  */
};
#else
static signed char keycode_to_oric[256];
#endif


/*
 *	keyboard_map_init - initialise keycode_to_oric;
 */
void keyboard_map_init (void)
{
  {
    int n;

    for (n = 0; n < sizeof keycode_to_oric / sizeof *keycode_to_oric; n++)
      keycode_to_oric[n] = -1;
  }

  keycode_to_oric[XKC_1]   = OK1;
  keycode_to_oric[XKC_2]   = OK2;
  keycode_to_oric[XKC_3]   = OK3;
  keycode_to_oric[XKC_4]   = OK4;
  keycode_to_oric[XKC_5]   = OK5;
  keycode_to_oric[XKC_6]   = OK6;
  keycode_to_oric[XKC_7]   = OK7;
  keycode_to_oric[XKC_8]   = OK8;
  keycode_to_oric[XKC_9]   = OK9;
  keycode_to_oric[XKC_0]   = OK0;
  keycode_to_oric[XKC_MIN] = OKMI;
  keycode_to_oric[XKC_EQ]  = OKEQ;
  keycode_to_oric[XKC_BSL] = OKBA;	/* On many keyboards it's 2 rows down */

  keycode_to_oric[XKC_TAB] = OKES;
  keycode_to_oric[XKC_Q]   = OKQ;
  keycode_to_oric[XKC_W]   = OKW;
  keycode_to_oric[XKC_E]   = OKE;
  keycode_to_oric[XKC_R]   = OKR;
  keycode_to_oric[XKC_T]   = OKT;
  keycode_to_oric[XKC_Y]   = OKY;
  keycode_to_oric[XKC_U]   = OKU;
  keycode_to_oric[XKC_I]   = OKI;
  keycode_to_oric[XKC_O]   = OKO;
  keycode_to_oric[XKC_P]   = OKP;
  keycode_to_oric[XKC_BRL] = OKBL;
  keycode_to_oric[XKC_BRR] = OKBR;
  keycode_to_oric[XKC_BS]  = OKDE;

  keycode_to_oric[XKC_CAP] = OKCT;
  keycode_to_oric[XKC_A]   = OKA;
  keycode_to_oric[XKC_S]   = OKS;
  keycode_to_oric[XKC_D]   = OKD;
  keycode_to_oric[XKC_F]   = OKF;
  keycode_to_oric[XKC_G]   = OKG;
  keycode_to_oric[XKC_H]   = OKH;
  keycode_to_oric[XKC_J]   = OKJ;
  keycode_to_oric[XKC_K]   = OKK;
  keycode_to_oric[XKC_L]   = OKL;
  keycode_to_oric[XKC_SEM] = OKSE;
  keycode_to_oric[XKC_APO] = OKAP;
  keycode_to_oric[XKC_ENT] = OKRE;

  keycode_to_oric[XKC_SHL] = OKLS;
  keycode_to_oric[XKC_LES] = OKLS;	/* Not on all keyboards */
  keycode_to_oric[XKC_Z]   = OKZ;
  keycode_to_oric[XKC_X]   = OKX;
  keycode_to_oric[XKC_C]   = OKC;
  keycode_to_oric[XKC_V]   = OKV;
  keycode_to_oric[XKC_B]   = OKB;
  keycode_to_oric[XKC_N]   = OKN;
  keycode_to_oric[XKC_M]   = OKM;
  keycode_to_oric[XKC_COM] = OKCO;
  keycode_to_oric[XKC_FUL] = OKFS;
  keycode_to_oric[XKC_SLA] = OKSL;
  keycode_to_oric[XKC_SHR] = OKRS;

  /* FIXME suited to 102-key keyboards (i.e. no windows keys) */
  keycode_to_oric[XKC_CTL] = OKLE;
  keycode_to_oric[XKC_ALL] = OKDO;
  keycode_to_oric[XKC_SPC] = OKSP;
  keycode_to_oric[XKC_ALR] = OKUP;
  keycode_to_oric[XKC_CTR] = OKRI;

  /* Also map the PC cursor keys */
  keycode_to_oric[XKC_UP]  = OKUP;
  keycode_to_oric[XKC_LEF] = OKLE;
  keycode_to_oric[XKC_DOW] = OKDO;
  keycode_to_oric[XKC_RIG] = OKRI;

  /* PC function key */
  keycode_to_oric[XKC_F1]  = -2;
  keycode_to_oric[XKC_F2]  = -2;
  keycode_to_oric[XKC_F3]  = -2;
  keycode_to_oric[XKC_F4]  = -2;
  keycode_to_oric[XKC_F5]  = -2;
  keycode_to_oric[XKC_F6]  = -2;
  keycode_to_oric[XKC_F7]  = -2;
  keycode_to_oric[XKC_F8]  = -2;
  keycode_to_oric[XKC_F9]  = -2;
  keycode_to_oric[XKC_F10] = -2;

  /* FIXME no PC key maps to Oric func */
}


void keyboard_dump (void)
{
  int n;

  for (n = 0; n < sizeof keycode_to_oric / sizeof *keycode_to_oric; n++)
  {
    if (keycode_to_oric[n] < 0)
      continue;
    printf ("0x%02x %03o\n", n, keycode_to_oric[n]);
  }
}


/*
 *	keyboard_handler - convert X11 keycode to Oric scan code
 */
void keyboard_handler (XEvent *ev)
{
  signed char oric_row_column;
  unsigned keycode = ev->xkey.keycode;

  if (ev->xany.type == KeyPress && keycode == 96)  /* F12 - screenshot */
  {
    screenshot ();
    return;
  }
  if (keycode < 0 || keycode > sizeof keycode_to_oric / sizeof *keycode_to_oric)
  {
    err ("bad keycode %d", (int) keycode);
    return;
  }
  oric_row_column = keycode_to_oric[keycode];
  if (oric_row_column == -2)	/* Function key */
  {
    if (ev->type == KeyPress)
      switch (keycode)
      {
	/* FIXME : missing: PrtScr */
	case 67 :				/* F1 - setup screen */
	  Sys_Request |= 0x8000;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	case 68 :				/* F2 - sound switch */
	  /* toggle_sound (); */
	  break;
	case 69 :				/* F3 - keyboard toggle */
	  /* toggle_keyboard (); */
	  break;
	case 70 :				/* F4 - double clock */
	  if (cycles < initial_cycles)
	    cycles = initial_cycles;
	  else if (cycles < LONG_MAX / 2)
	    cycles *= 2;
	  break;
	case 71 :				/* F5 - halve clock */
	  if (cycles > initial_cycles)
	    cycles = initial_cycles;
	  else if (2 * (cycles / 2) == cycles)
	    cycles /= 2;
	  break;
	case 72 :				/* F6 - Jasmin boot */
	  Sys_Request |= 0x200;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	case 73 :				/* F7 - NMI */
	  Sys_Request |= 0x400;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	case 74 :				/* F8 - power on reset */
	  Sys_Request |= 0x800;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	case 75 :				/* F9 - dump */
	  Sys_Request |= 0x1000;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	case 76 :				/* F10 - quit */
	  Sys_Request |= 0x2000;
	  if (env_screen == 1)
	    exit_menu = 1;
	  break;
	default :				/* Can't happen */
	  err ("bad fn keycode %d\n", (int) keycode);
      }
    return;
  }
  else if (oric_row_column == -1)
  {
    return;
  }
  else
  {
    if (ev->type == KeyPress)
      Kbd_Matrix[oric_row_column & 7] |= 1 << (oric_row_column >> 3);
    else if (ev->type == KeyRelease)
      Kbd_Matrix[oric_row_column & 7] &= ~ (1 << (oric_row_column >> 3));
    return;
  }
}


void disable_keypad (void)
{
  ;					/* FIXME */
}

void enable_keypad (void)
{
  ;					/* FIXME */
}


void Read_Joystick_on_printer_port (void)
{
  ;					/* FIXME */
}


void Read_Joystick_on_VIA2 (void)
{
  ;					/* FIXME */
}

