/*
 *	oric.c - main
 *	FF sometime in 1994-1997
 */

/*
Parts of this file copyright 1994-1997 Fabrice Francs.
Parts of this file copyright 2000-2003 Andr Majorel.

This program is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published by the
Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "xeuphoric.h"
#include "dsp.h"
#include "screenshot.h"
#include "stats.h"

/* FIXME use headers */
extern int init(char argc, char *argv[]);
extern void setup(char *env[]);
#ifdef X11
extern void x11_start(void);
extern void x11_end(void);
#else 
extern void vga_start(void);
extern void vga_end(void);
#endif
extern void adlib_init(void);
extern void hook_interrupts(void);
extern void redirect_output(void);
extern void delete_output(void);
extern void init_time(void);
extern void Restart(void);
extern void Reset(void);
extern void restore_time(void);
extern void restore_interrupts(void);
extern void adlib_end(void);
extern char restart_flag;

int main(int argc, char *argv[], char *env[])
{
	stats_init();
	if (argc>=2 && strcmp(argv[1],"--help")==0) {
		puts(
"xeuphoric - an Oric emulator\n"
"http://www.teaser.fr/~amajorel/xeuphoric/\n"
"Usage:\n"
"  xeuphoric --help\n"
"  xeuphoric --version\n"
"  xeuphoric [-1|a|t] [-djJpQr] [-A file] [-s num] [-S arts|auto|oss]\n"
"    [-X always|auto|never] [-z num] file ...\n"
"Options:\n"
"  --help     print usage on standard output and exit successfully\n"
"  --version  print version number on standard output and exit successfully\n"
"  -1         emulate an Oric-1\n"
"  -a         emulate an Oric Atmos (this is the default)\n"
"  -A file    pathname of OSS audio device (default /dev/dsp)\n"
"  -d         emulate a Microdisk floppy disk drive\n"
"  -j         emulate a Jasmin floppy disk drive\n"
"  -J         emulate a Telestrat joystick\n"
"  -p         emulate a P.A.S.E. joystick interface\n"
"  -Q         quit on illegal instructions\n"
"  -r         restart from a dump\n"
"  -s num     emulate at num percent of the normal speed (default 100)\n"
"  -S arts|auto|oss  sound output method (default is auto)\n"
"  -t         emulate a Telestrat\n"
"  -X always|auto|never  XShm usage policy (default is auto)\n"
"  -z num     set zoom factor to num (1, 2 or 3, default 1)");
		exit(0);
	}
	if (argc>=2 && strcmp(argv[1],"--version")==0) {
		puts(version);
		if (fflush(stdout))
			exit(1);
		exit(0);
	}
#ifdef SVGALIB
	if (geteuid()) {
		err("Euphoric needs to be run 'root setuid' (see manual)");
		exit(1);
	}
	get_rights();
#endif
	setup(env);
	keyboard_map_init();
	/*keyboard_dump();*/
	restart_flag=init(argc,argv);
#ifdef X11
	x11_start();
#else
	vga_start();
#endif
/*
	adlib_init();
*/
	sound_open();
	hook_interrupts();
	init_time();
	redirect_output();
	if (restart_flag) Restart();
	else Reset();
	return 0;			/* NOT REACHED */
}

void Quit(void)
{
	restore_time();
/*
	delete_output();
*/
	restore_interrupts();
#ifdef X11
	x11_end();
#else
	vga_end();
#endif
/*
	adlib_end();
*/
	if (sound)
	  sound_close();
	stats_dump(stderr);		/* stdout does not work */
	exit(0);
}

