/*
 *	x11render1bpp.c - render an Oric frame into a 1-BPP buffer
 *	AYM 2002-01-27
 */

/*
This file is copyright Andr Majorel 2003.

This program is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published by the
Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307, USA.
*/


/* This file is not supposed to be compiled directly. It is
   included from x11.c.
   
   Assuming :

     - img->bitmap_unit <= CHAR_BIT,

     - the significant bit in colours[] is b0.
*/


const char *b;

const size_t bytes_per_line  = img->bytes_per_line;
const size_t pixels_per_byte = img->bitmap_unit;
unsigned char *to  = (unsigned char *) img->data + 0 * bytes_per_line;
unsigned char *to2 = (unsigned char *) img->data + 1 * bytes_per_line;
unsigned char *to3 = (unsigned char *) img->data + 2 * bytes_per_line;

int y;

b = buf;
for (y = 0; y < VIDEO_HEIGHT; y++)
{
  const char *bmax;

  if (b[VIDEO_WIDTH])
  {
    if (y0 < 0)
      y0 = y;
    y1 = y;

    /* Render one scan line */
    if (zoom == 1)
    {
      int tobit = 0;
      unsigned char *to0 = to;

      for (bmax = b + VIDEO_WIDTH; b < bmax; b++)
      {
	if (tobit % pixels_per_byte == 0)
	  *to = 0;

#if STRAIGHT_BITMAP
	*to <<= 1;
	*to |= colours[*b];
#else
	*to >>= 1;
	*to |= (colours[*b] << (pixels_per_byte - 1));
#endif
	tobit++;
	if (tobit % pixels_per_byte == 0)
	  to++;
      }
      to = to0 + bytes_per_line;
    }
    else if (zoom == 2)
    {
      int tobit = 0;
      unsigned char *to0  = to;
      unsigned char *to02 = to2;

      for (bmax = b + VIDEO_WIDTH; b < bmax; b++)
      {
	if (tobit % CHAR_BIT == 0)
	{
	  *to  = 0;
	  *to2 = 0;
	}

#if STRAIGHT_BITMAP
	*to  <<= 1;
	*to  |= colours[*b];
	*to  <<= 1;
	*to  |= colours[*b];
	*to2 <<= 1;
	*to2 |= colours[*b];
	*to2 <<= 1;
	*to2 |= colours[*b];
#else
	*to  >>= 1;
	*to  |= (colours[*b] << (CHAR_BIT - 1));
	*to  >>= 1;
	*to  |= (colours[*b] << (CHAR_BIT - 1));
	*to2 >>= 1;
	*to2 |= (colours[*b] << (CHAR_BIT - 1));
	*to2 >>= 1;
	*to2 |= (colours[*b] << (CHAR_BIT - 1));
#endif
	tobit += 2;
	if (tobit % pixels_per_byte == 0)
	{
	  to++;
	  to2++;
	}
      }
      to  = to0  + 2 * bytes_per_line;
      to2 = to02 + 2 * bytes_per_line;
    }
    else if (zoom == 3)
    {
      int tobit = 0;
      int horizpix = 0;
      unsigned char *to0  = to;
      unsigned char *to02 = to2;
      unsigned char *to03 = to3;

      for (bmax = b + VIDEO_WIDTH; b < bmax; b++)
      {
	for (horizpix = 0; horizpix < zoom; horizpix++)
	{
	  if (tobit % CHAR_BIT == 0)
	  {
	    *to  = 0;
	    *to2 = 0;
	    *to3 = 0;
	  }
#if STRAIGHT_BITMAP
	  *to  <<= 1;
	  *to  |= colours[*b];
	  *to2 <<= 1;
	  *to2 |= colours[*b];
	  *to3 <<= 1;
	  *to3 |= colours[*b];
#else
	  *to  >>= 1;
	  *to  |= (colours[*b] << (CHAR_BIT - 1));
	  *to2 >>= 1;
	  *to2 |= (colours[*b] << (CHAR_BIT - 1));
	  *to3 >>= 1;
	  *to3 |= (colours[*b] << (CHAR_BIT - 1));
#endif
	  tobit++;
	  if (tobit % pixels_per_byte == 0)
	  {
	    to++;
	    to2++;
	    to3++;
	  }
	}
      }
      to  = to0  + 3 * bytes_per_line;
      to2 = to02 + 3 * bytes_per_line;
      to3 = to03 + 3 * bytes_per_line;
    }
    b++;				/* KLUDGE - skip over the boolean */
  }
  else
  {
    b += VIDEO_WIDTH + 1;		/* KLUDGE - +1 to skip over the bool */
    if (zoom == 1)
    {
      to += bytes_per_line;
    }
    else if (zoom == 2)
    {
      to  += 2 * bytes_per_line;
      to2 += 2 * bytes_per_line;
    }
    else if (zoom == 3)
    {
      to  += 3 * bytes_per_line;
      to2 += 3 * bytes_per_line;
      to3 += 3 * bytes_per_line;
    }
  }
}

/* Do the trace thing */
#if 0
{
  int n = tracehead;
  ximage_pixel_t *tomax = to + VIDEO_WIDTH;
  for (; to < tomax; to++)
  {
    if (n >= sizeof tracebuf / sizeof *tracebuf)
      n = 0;
    if (tracebuf[n])
      *to = colours[4];
    else
      *to = colours[3];
    n++;
  }
}
#endif
